/*
 * PgGraph.java
 *
 * Created on December 5, 2004, 6:17 PM
 */

package Grapher.geom;

import java.awt.Color;

import jv.function.PuFunction;
import jv.number.PuDouble;
import jv.object.PsPanel;
import jv.geom.PgElementSet;
import jv.number.PdColor;
import jv.number.PuInteger;
import jv.object.PsDebug;
import Grapher.util.PuDomain;
import Grapher.util.PuParams;
import Grapher.Viewer.PjViewer;
import jv.object.PsUpdateIf;

/**
 * Geometry class defining graphs of the form z=f(x,y)
 *
 * @author David
 * @version 1.0.281005
 * @see PgElementSet
 */
public class PgGraph extends PgElementSet {

        protected       PuFunction                      m_func2;
        protected       PuDomain                        m_domain;
        protected       PuParams                        m_params;
        protected       boolean                         m_bHasVf = false;
        protected       boolean                         m_doanim = false;
        protected       PjViewer                        m_PjViewer;
        
        /** Constructor.  Calls init()
         * @param   PName   Name used to reference graph 
         */
	public PgGraph(String PName) {
            
	    if (getClass() == PgGraph.class)
		  init(PName);
	}
        
        /** Initializes the Functions and domain.  Initializes color parameters 
         * @param PName Name used to reference graph
         */
	public void init(String PName) {
               
            super.init();
            setName(PName);

                    
  
                    setGlobalEdgeColor(Color.black);
                    
                    m_func2 = new PuFunction(2,1);
                    m_func2.setName("Functions");
                    m_func2.setVariables(new String[] {"x","y"});
                    m_func2.setExpressionNames(new String[] {"z"});
                    m_func2.setExpressions(new String[] {"x^2+y^2"});
                    m_func2.addParameter("time",1.0);
                    m_func2.setParent(this);
                    
                    m_domain = new PuDomain();
                    m_domain.setVars(new String[] {"x","y"});
                    m_domain.setAllowSlices(true);
                    m_domain.init();
                    m_domain.setParent(this);
                    
                    m_params  = new PuParams();
                    m_params.init(3);
                    m_params.registerWith(m_func2);
                    m_params.setParent(this);
                    
                    assureElementColors();
                    showElementColors(true);
                    showVertices(false);
                    
		
 	}
        
        /** Does nothing */
	public void start() {
		if (PsDebug.NOTIFY) PsDebug.notify("start: ");

	}
        
        public void setParent(PsUpdateIf parent) {
		m_PjViewer = (PjViewer)parent;
                computeSurf();
        }
	/**
	 * Update the class whenever a child has changed.
	 * Method is usually invoked from the children.
	 */
	public boolean update(Object event) {
		if (event == this) {
                        if (m_doanim) {
                            computeSurf();
                        }
			super.update(this);
                        return true;
		} else if (event == m_domain) {
                        computeSurf();
                        
                        return super.update(this);
                } else if (event == m_func2) {
                        if ((m_func2.getExpression(0).indexOf("time")>-1)){
                            m_doanim = true;
                        } else {
                            m_doanim = false;
                        }
                        
                        computeSurf();
			return super.update(this);
                } else {
                    return super.update(this);
                }
                
		
	}
        
        /** Main computation method.  Computes vertices and creates elements */
	public synchronized void computeSurf() {
                
                m_func2.setParameter("time",m_PjViewer.getTime());
                int ulines, vlines;
                
                double umin;
                double vmin;
                double x,y,z;
         
                umin = m_domain.getMin(0);
                vmin = m_domain.getMin(1);
                double u= umin;
                double v= vmin;
                double deltaU, deltaV;
                
                if (m_domain.isSliced(0)){
                   ulines = 2;
                   deltaU = .1;
                } else {
                    ulines = m_domain.getDiscr(0);
                    deltaU	= (m_domain.getMax(0)-u)/(ulines-1.);
                
                }
                
                if (m_domain.isSliced(1)){
                    vlines = 2;
                    deltaV = .1;
                } else {
                    vlines = m_domain.getDiscr(1);
                    deltaV = (m_domain.getMax(1)-v)/(vlines-1.);
                }
                
                int numVertices = ulines*vlines;
                if (numVertices == 1)
			return;
		setNumVertices(numVertices);
               
                int ind = 0;
		for (int i=0; i<ulines; i++) {
			for (int j=0; j<vlines; j++) {
                 
                                double[] inp = {u,v};
                                
                               x = u;
                               y = v;
                                z = m_func2.eval(0,inp);
                                
                                setVertex(ind,x,y,z);
						
				v += deltaV;
                                ind++;
			}
                        v = vmin;
                        u += deltaU;
		}
		// Generate connectivity for quadrilateral mesh
		makeQuadrConn(ulines, vlines);

		// Compute surface normals
		makeElementNormals();
		makeVertexNormals();
		
                //Make colors based on location in xyz grid
                showEdges(m_domain.isShowEdges());
                int colorby = m_domain.getColorBy();
                if (colorby==0){
                    makeElementColorsFromXYZ();
                } else if (colorby==1) {
                    makeElementColorsFromZHue();
                }
            
	}
        
        
    
}
