/*
 * PSurf.java
 *
 * Created on December 5, 2004, 6:14 PM
 */

package Grapher.geom;

import jv.function.PuFunction;
import jv.number.PuDouble;
import jv.object.PsPanel;
import jv.geom.PgElementSet;
import jv.number.PdColor;
import jv.number.PuInteger;
import jv.object.PsDebug;
import Grapher.util.*;
import jv.object.PsUpdateIf;
import Grapher.Viewer.PjViewer;

/**
 * Geometry class for Parametrically defined surfaces
 *
 * @author David Lippman
 * @version 1.0.122204
 * @see PgElementSet
 */
public class PgPSurf extends PgElementSet {

    protected       PuFunction                      m_func2;
    protected       PuDomain                        m_domain;
    protected       PuParams                        m_params;
    protected       boolean                         m_bHasVf = false;  
    protected       boolean                         m_doanim = false;
    protected       PjViewer                        m_PjViewer;
    
    /** Creates a new instance of PSurf 
     * @param Pname Name used to reference graph
     */
    public PgPSurf(String Pname) {
        super(3);
        if (getClass() == PgPSurf.class) {
            init(Pname);
        }
    }
    
    /** Initializes the domain and function
     * @param Pname Name used to reference graph
     */
    public void init(String Pname) {
        super.init();
        setName(Pname);
        
            
        m_func2 = new PuFunction(2,3);
        m_func2.setName("Functions");
        m_func2.setVariables(new String[] {"u","v"});
        m_func2.setExpressionNames(new String[] {"x","y","z"});
        m_func2.setExpressions(new String[] {"u*cos(v)","u*sin(v)","v"});
        m_func2.addParameter("time",1.0);
        m_func2.setParent(this);
        
        m_domain = new PuDomain();
        m_domain.setVars(new String[] {"u","v"});
        m_domain.setDefVals(new double[] {0,2*Math.PI});
        m_domain.setAllowSlices(true);
        m_domain.init();
        m_domain.setParent(this);

        m_params  = new PuParams();
        m_params.init(3);
        m_params.registerWith(m_func2);
        m_params.setParent(this);
        
        assureElementColors();
        showElementColors(true);
        showVertices(false);
        
        
    }
    
    /** Main geometry method.  Creates vertices and elements */
    public synchronized void computeSurf() {
            
                int ulines, vlines;
                             
                double umin;
                double vmin;
                double x,y,z;
         
                umin = m_domain.getMin(0);
                vmin = m_domain.getMin(1);
                double u= umin;
                double v= vmin;
                double deltaU, deltaV;
                
                m_func2.setParameter("time",m_PjViewer.getTime());
                
                if (m_domain.isSliced(0)){
                   ulines = 2;
                   deltaU = .1;
                } else {
                    ulines = m_domain.getDiscr(0);
                    deltaU = (m_domain.getMax(0)-u)/(ulines-1.);
                
                }
                
                if (m_domain.isSliced(1)){
                    vlines = 2;
                    deltaV = .1;
                } else {
                    vlines = m_domain.getDiscr(1);
                    deltaV = (m_domain.getMax(1)-v)/(vlines-1.);
                }
                
                int numVertices = ulines*vlines;
                if (numVertices == 1)
			return;
		setNumVertices(numVertices);
               
                int ind = 0;
		for (int i=0; i<ulines; i++) {
			for (int j=0; j<vlines; j++) {
                 
                                double[] inp = {u,v};
                                
                                
                                    x = m_func2.eval(0,inp);
                                    y = m_func2.eval(1,inp);
                                    z = m_func2.eval(2,inp);
                                
                                setVertex(ind,x,y,z);
						
				v += deltaV;
                                ind++;
			}
                        v = vmin;
                        u += deltaU;
		}
		// Generate connectivity for quadrilateral mesh
		makeQuadrConn(ulines, vlines);

		// Compute surface normals
		makeElementNormals();
		makeVertexNormals();
		
		showEdges(m_domain.isShowEdges());
                int colorby = m_domain.getColorBy();
                if (colorby==0){
                    makeElementColorsFromXYZ();
                } else if (colorby==1) {
                    makeElementColorsFromZHue();
                }
            
	}
    
    /** does nothing */
    public void start() {
		if (PsDebug.NOTIFY) PsDebug.notify("start: ");
    }
    
    public void setParent(PsUpdateIf parent) {
		m_PjViewer = (PjViewer)parent;
                computeSurf();
        }
    
    /** Handles update calls.  Recomputes surface when domain or function change */
    public boolean update(Object event) {
		if (event == this) {
                     if (m_doanim) {
                            computeSurf();
                        }
			super.update(this);
                        return true;
		} else if (event == m_domain) {
                    computeSurf();
                    return super.update(this);
                    
                } else if (event == m_func2) {
                    if ((m_func2.getExpression(0).indexOf("time")>-1)||(m_func2.getExpression(1).indexOf("time")>-1)||(m_func2.getExpression(2).indexOf("time")>-1)){
                            m_doanim = true;
                        } else {
                            m_doanim = false;
                        }
                    computeSurf();
                    return super.update(this);
                }
                return true;
    }
}
