/*
 * PgVectorField.java
 *
 * Created on December 5, 2004, 6:20 PM
 */

package Grapher.geom;

import java.awt.Color;
import jv.function.PuFunction;
import jv.number.PuDouble;
import jv.object.PsPanel;
import jv.geom.PgPointSet;
import jv.number.PdColor;
import jv.number.PuInteger;
import jv.object.PsDebug;
import jv.vecmath.PdVector;
import jv.vecmath.PiVector;
import jv.geom.PgVectorField;
import Grapher.util.*;
import jv.object.PsUpdateIf;
import Grapher.Viewer.PjViewer;
/**
 * Geometry class for Vector fields on a grid
 *
 * @author David Lippman
 * @version 1.0.122204
 * @see PgPointSet, PgVectorField
 */

public class PgVecFld extends PgPointSet {
    protected       PuFunction                      m_func2;
    protected       PuDomain                        m_domain;
    protected       PuParams                        m_params;
    protected       PgVectorField                   m_vf;
    protected       boolean                         m_doanim = false;
    protected       PjViewer                        m_PjViewer;
    
    /** Creates a new instance of PgVectorField 
     * @param   Pname   Name used to reference graph
     */
    public PgVecFld(String Pname) {
        super(3);
        if (getClass() == PgVecFld.class) {
            init(Pname);
        }
    }
    
    /** Initializes the function and domain and sets color parameters
     *  @param Pname    Name used to reference graph
     */
    public void init(String Pname) {
        super.init();
        setName(Pname);
        
        m_vf = new PgVectorField(3, PgVectorField.VERTEX_BASED);
        
        m_func2 = new PuFunction(3,3);
        m_func2.setName("Functions");
        m_func2.setVariables(new String[] {"x","y","z"});
        m_func2.setExpressionNames(new String[] {"dx","dy","dz"});
        m_func2.setExpressions(new String[] {"-y","x","z"});
        m_func2.addParameter("time",1.0);
        m_func2.setParent(this);
        
        m_domain = new PuDomain();
        m_domain.setVars(new String[] {"x","y","z"});
        m_domain.setDefVals(new double[] {-5.0,5.0});
        m_domain.setDefDiscr(new int[] {7});
        m_domain.setAllowSlices(true);
        m_domain.setIsVF(true);
        m_domain.init();
        m_domain.setParent(this);
        
        m_params  = new PuParams();
        m_params.init(3);
        m_params.registerWith(m_func2);
        m_params.setParent(this);
        
        assureVertexColors();
        showVertexColors(true);
        showVertices(false);
        m_vf.assureVectorColors();
                    m_vf.showIndividualMaterial(true);
                    m_vf.showVectorArrows(true);
                    m_vf.setGlobalVectorColor(Color.black);
        
        
        
        
       // computeVf();
        
        m_vf.setGeometry(this); //Set vectorfield to use this class (Pointset) as geometry to attatch to
        addVectorField(m_vf); //Tell this class (pointset) to use the vectorfield
        
        
        
    }
    
    /** does nothing */
    public void start() {
		if (PsDebug.NOTIFY) PsDebug.notify("start: ");
    }
    
    public void setParent(PsUpdateIf parent) {
		m_PjViewer = (PjViewer)parent;
                computeVf();
        }

    
    /** Handles update calls.  Recomputes vector field when domain or function change */
    public boolean update(Object event) {
		if (event == this) {
                    if (m_doanim) {
                            computeVf();
                        }
			super.update(this);
                        return true;
		} else if (event == m_domain) {
                    computeVf();
                    return super.update(this);
                    
                } else if (event == m_func2) {
                    if ((m_func2.getExpression(0).indexOf("time")>-1)||(m_func2.getExpression(1).indexOf("time")>-1)||(m_func2.getExpression(2).indexOf("time")>-1)){
                            m_doanim = true;
                        } else {
                            m_doanim = false;
                        }
                    computeVf();
                    return super.update(this);
                }
                return true;
    }
    
    /** Creates a grid of vertices, and creates vectors at those points */
    public synchronized void computeVf() {
 
            int[] discr;
            double[] delta;
            delta = new double[3];
            discr = m_domain.getDiscr();
            
            m_func2.setParameter("time",m_PjViewer.getTime());
            
            for (int i=0; i<3;i++) {
                if (m_domain.isSliced(i)) {
                    discr[i] = 1;
                    delta[i] = .1;
                } else {
                    delta[i] = (m_domain.getMax(i)-m_domain.getMin(i))/(discr[i]-1.);
                }
            }
            
            int numVertices = discr[0]*discr[1]*discr[2];
            setNumVertices(numVertices);
            m_vf.setNumVectors(numVertices);
            
            int i,j,k;
            double u,v,w;
            
            int ind = 0;
  
                for (i=0; i<discr[0];i++){
                    for (j=0; j<discr[1];j++){
                        for (k=0;k<discr[2];k++){
                                u = m_domain.getMin(0)+delta[0]*i;
                                v = m_domain.getMin(1)+delta[1]*j;
                                w = m_domain.getMin(2)+delta[2]*k;
                                double[] inp = {u,v,w};
				double x,y,z;
                                x = m_func2.eval(0,inp);
                                y = m_func2.eval(1,inp);
                                z = m_func2.eval(2,inp);
                                
                                double scl;
                                if (m_domain.isSclVecs()) {
                                    double length = Math.sqrt(x*x+y*y+z*z);
                                    scl = m_domain.getVeclength()/(3.*length+0.1);
                                } else {
                                    scl = m_domain.getVeclength()/10.;
                                }
 //                               double length = Math.sqrt(x*x+y*y+z*z);
 //                               double scl = Math.min(deltaU,Math.min(deltaV,deltaW))/(length+0.1);
                      
                                setVertex(ind,u,v,w);
                                m_vf.setVector(ind,x*scl,y*scl,z*scl);
                                ind++;
                        }
                    }
                }
         
        }
}
