/*
 * PuDomain.java
 *
 * Created on December 5, 2004, 6:55 PM
 */

package Grapher.util;

import jv.object.PsObject;
import jv.function.PuFunction;
import jv.number.PuDouble;
import jv.number.PdColor;
import jv.number.PuInteger;
import jv.object.PsDebug;
import jv.vecmath.PdVector;
import jv.vecmath.PiVector;
/**
 * Domain class.  Holds domain information for any geometry.
 *
 * @author David Lippman
 * @version 1.0.122204
 */
public class PuDomain extends PsObject {
        protected       double[]                        m_defVals = {-5.0,5.0};
        protected       double[]                        m_defBnds = {-20.0,20.0,.1,1.};
        protected       int[]                           m_defDiscr = {20};
        protected       int                             m_type;
        protected       int                             numvariables;
        protected       int                             m_numDiscr = -1;
        protected       boolean                         m_bIsVF = false;
        protected       boolean                         m_bAllowSlices = false;
        protected       String[]                        m_varNames = {"x","y"};
        protected       PuDouble[][]                    m_var;
        protected       double[]                        m_varValBU;
        protected       PuInteger[]                     m_discr;
        protected       PuDouble                        m_veclength;
        protected       boolean                         m_bSclVecs = false;
        protected       boolean[]                       m_bShowSlice;
        protected       boolean                         m_bShowEdges = true;
        protected       int                             m_ColorBy = 0;
                                                        //0: by XYZ, 1: by Z

        
    /** Creates a new instance of PuDomain */
    public PuDomain() {
        super();
        if (getClass() == PuDomain.class) {
            
        }
    }
    
    /** Initializes the domain.  Call after setting variables and any parameters */
    public void init() {
        super.init();
        
       	numvariables = m_varNames.length;
        m_var = new PuDouble[numvariables][2];
        m_varValBU = new double[numvariables];
        
//        m_discr = new PuInteger[m_bOneDiscr? 1 : numvariables];
        // If default -1, create discr for each variable.  Otherwise, set number requested
        if (m_numDiscr == -1) {
            m_discr = new PuInteger[numvariables];
        } else {
            m_discr = new PuInteger[m_numDiscr];
        }
        
        m_bShowSlice = new boolean[numvariables];
        
        for (int i=0; i<numvariables;i++) {
            //create min
            m_var[i][0] = new PuDouble(m_varNames[i] + " min", this);
            m_var[i][0].setDefBounds(m_defBnds[0],m_defBnds[1],m_defBnds[2],m_defBnds[3]);
            if (m_defVals.length == 2) {
                m_var[i][0].setDefValue(m_defVals[0]);
            } else {
                m_var[i][0].setDefValue(m_defVals[2*i]);
            }
            //create max
            m_var[i][0].init();
            m_var[i][1] = new PuDouble(m_varNames[i] + " max", this);
            m_var[i][1].setDefBounds(m_defBnds[0],m_defBnds[1],m_defBnds[2],m_defBnds[3]);
            if (m_defVals.length == 2) {
                m_var[i][1].setDefValue(m_defVals[1]);
            } else {
                m_var[i][1].setDefValue(m_defVals[2*i+1]);
            }
            m_var[i][1].init();
            //If default -1, create discr for each variable. Otherwise create number requested
            if (m_numDiscr == -1) {
//            if (!m_bOneDiscr) {
                m_discr[i] = new PuInteger(m_varNames[i] + " discr", this);
                m_discr[i].setDefBounds(2,300,1,5);
                if (m_defDiscr.length == 1) {
                     m_discr[i].setDefValue(m_defDiscr[0]);
                } else {
                    m_discr[i].setDefValue(m_defDiscr[i]);
                }
                m_discr[i].init();
            }
            m_bShowSlice[i] = false;
        }
        if (m_numDiscr ==1) {
//        if (m_bOneDiscr) {
            m_discr[0] = new PuInteger("discr", this);
            m_discr[0].setDefBounds(2,300,1,5);
            m_discr[0].setDefValue(m_defDiscr[0]);
            m_discr[0].init();
           
        }
        
        // If is a vector field, create vector length
        if (m_bIsVF) {
            m_veclength  = new PuDouble("Vector Length",this);
            m_veclength.setDefBounds(0,4,0.1,0.5);
            m_veclength.setDefValue(1);
            m_veclength.init();
        }
        
    }
    
    /** 
     *  Sets starting values for variables.  
     *  If not set, defaults to {-5.0,5.0}
     *  <p>
     *  Examples: setDefVals(new double[] {-3.0,3.0})  Sets all variables min=-3.0, max=3.0<BR>
     *            setDefVals(new double[] {0.0,5.0,-2.0,2.0}   Sets first var min=0, max=5, second var min=-2, max=2
     *
     *  @param vals array of doubles giving min and max for each variable.  If two elements, same values are used for all variables.
     */
    public void setDefVals(double[] vals) {
        m_defVals = vals;
    }
    
    /** Sets bounds for all domain variable sliders<BR>
     *  If not set, defaults to {-20.0,20.0,.1,1.}
     *
     *  @param bnds array of doubles {min, max, smallincrement, pageincrement}
     */
    public void setDefBnds(double[] bnds) {
        m_defBnds = bnds;
    }
    
    /** Sets discretizations<BR>
     *  If not set, defaults to {20}<BR>
     *  If array has one entry, it is used for all discrs.  Otherwise, provide an entry for each variable
     *
     *  @param  discr   array of ints - discritization values.  
     */
    public void setDefDiscr(int[] discr) {
        m_defDiscr = discr;
    }
    
    /** Set variable names<BR>
     *  If not set, defaults to {"x","y"}
     *
     *  @param  vars    String array of domain variable names
     */
    public void setVars(String[] vars) {
        m_varNames = vars;
    }
    
    /** Provide checkboxes for doing variable slices<BR>
     *  Defaults to false
     *
     *  @param  allow   True to allow slices
     */
    public void setAllowSlices(boolean allow) {
        m_bAllowSlices = allow;
    }
    
    
   
    /** Sets number of discritizations<BR>
     *  Defaults to -1
     *  
     * @param   num   Set =0 for no Discr, 1 for 1 Discr, -1 for Discr for each variable 
     */
    public void setNumDiscr(int num) {
        m_numDiscr = num;
    }
    
    /** Sets whether domain is for a vector field.  If set, will provide a Vector Length slider<BR>
     *  Defaults to false
     *
     * @param   is  True if it is vector field
     */
    public void setIsVF(boolean is) {
        m_bIsVF = is;
    }
    
    /** Check to see if Scale Vectors to One Length box is checked<BR>
     *  Only used if setIsVF(true) has been called
     *
     *  @return True if Scale Vectors has been checked
     */
    public boolean isSclVecs() {
        return m_bSclVecs;
    }
    
    /** Get value of Vector length slider
     *
     * @return  double value of vector length slider
     */
    public double getVeclength() {
        return m_veclength.getValue();
    }
    
    /** Get value of a discritization
     *
     * @param   i   variable you want the discr for
     * @return  int value of discr
     */
    public int getDiscr(int i) {
        return m_discr[i].getValue();
    }
    
    /** Get values of discritizations
     *  Should only be used if you have a discr for each variable
     *
     * @return  int array of discr values
     */
    public int[] getDiscr() {
        int[] discr;
        discr = new int[numvariables];
        for (int i=0; i<numvariables; i++) {
            discr[i] = m_discr[i].getValue();
        }
        return discr;
    }
    
    /** Gets the domain minimum for a variable
     *
     * @param   i   Variable you want the min domain for
     * @return  double value of domain min
     */
    public double getMin(int i) {
        return m_var[i][0].getValue();
    }
    
     /** Gets the domain maximum for a variable
     *
     * @param   i   Variable you want the max domain for
     * @return  double value of domain max
     */
    public double getMax(int i) {
        return m_var[i][1].getValue();
    }
    
    /** Checks whether a variable Slice checkbox is checked
     *
     *  @param  i   variable to check if its sliced
     *  @return true if that slice checkbox is checked for that variable
     */
    public boolean isSliced(int i) {
        return m_bShowSlice[i];
    }
    
    /** Checks whether Show Edges checkbox is checked
     *
     * @return true if show edges checkbox is checked
     */
    public boolean isShowEdges() {
        return m_bShowEdges;
    }
    
    /** Gets Color By setting
     *
     * @return 0 for by XYZ, 1 for by Z
     */
    public int getColorBy() {
        return m_ColorBy;
    }
    
    /** Handles update events.  Called when domain sliders are changed */
    public boolean update(Object event) {
                if (event == this) {
                        super.update(this);
                        return true;
		} else if (event == m_discr) {
                        
			update(this);
			return true;
		} 
                
                for (int i=0; i<numvariables;i++) {
                    if (event == m_var[i][0]) {
                        if (m_var[i][0].getValue() >= m_var[i][1].getValue()) {
                            m_var[i][0].setValue(m_var[i][1].getValue()-.01);
                        }
                    } else if (event == m_var[i][1]) {
                        if (m_var[i][0].getValue() >= m_var[i][1].getValue()) {
                            m_var[i][1].setValue(m_var[i][0].getValue()+.01);
                        }
                    }
                }
                
                super.update(this);
                
		return true;
	}
    
        
    
}
